import asyncio
import uuid
from threading import Thread

from a2a.client import ClientFactory, ClientConfig
from a2a.types import (
    AgentCard,
    PushNotificationConfig,
    Message,
    Role,
    TextPart,
    Task,
)

from a2a_acl.a2a_utils.card_holder import download_card


def extract_text_from_message(m: Message) -> str:
    return m.parts[0].root.text


def extract_text_from_task(t: Task) -> str:
    res = ""
    for a in t.artifacts:
        for p in a.parts:
            if p.root.kind == "text":
                res = res + p.root.text
    return res


def default_message_handler(m: Message) -> None:
    print("(Sync reply received (Message): " + extract_text_from_message(m) + ")")


def default_task_handler(t: Task) -> None:
    print("(Sync reply received (Task Artifact): " + extract_text_from_task(t) + ")")


class SendFailureException(Exception):
    pass


# derived from https://github.com/a2aproject/a2a-samples/blob/main/samples/python/agents/number_guessing_game/utils/protocol_wrappers.py
async def send_acl_message(
    target: AgentCard | str,
    illocution: str,
    text: str,
    my_url: str,
    codec: str,
    message_processor=default_message_handler,
    task_processor=default_task_handler,
):
    """Send *text* to the target agent via the A2A ``message/send`` operation.

    Args:
        target: card or url of the target agent.
        illocution : illocution such as 'tell' or 'achieve'
        text: Payload to send as a plain-text message.
        my_url: the url of sender's server for replies.
        codec: the codec used to code the content of the message.
        message_processor: the processor used to process the 'message' text parts.
        task_processor: the processor used to process the 'task' text parts.

    Returns:
        Union[Task, Message]: The final object produced by the agent—normally a
        ``Task`` but may be a plain ``Message`` for very small interactions.
    """
    if isinstance(target, str):
        target = await download_card(target)
    _client_factory = ClientFactory(
        ClientConfig(push_notification_configs=[PushNotificationConfig(url=my_url)])
    )
    client = _client_factory.create(target)
    md: dict[str, str] = {"illocution": illocution, "codec": codec}
    msg = Message(
        kind="message",
        role=Role.agent,
        message_id=uuid.uuid4().hex,
        parts=[TextPart(text=text, metadata=md)],
        task_id=None,
        metadata=md,
    )

    last_message = None

    try:
        async for event in client.send_message(msg):
            # Unwrap tuple from transport implementations
            if isinstance(event, tuple):
                (event, other) = event
                if other is not None:
                    print(
                        "Additional info received with "
                        + str(type(event))
                        + ":"
                        + str(other)
                    )
            if isinstance(event, Message):
                if message_processor is not None:
                    message_processor(event)
                last_message = event
            elif isinstance(event, Task):
                if task_processor is not None:
                    task_processor(event)
            else:
                print("Event not supported: " + str(event))

        if last_message is not None:
            return last_message
        else:
            return None
    except Exception as e:
        print("Send failed.")
        raise SendFailureException


class SendMessageThread(Thread):
    def __init__(
        self,
        dest: AgentCard | str,
        illocution: str,
        content: str,
        sender: str,
        codec: str,
        message_processor=default_message_handler,
        task_processor=default_task_handler,
    ):
        super().__init__()
        self.dest = dest
        self.illocution = illocution
        self.content = content
        self.reply_to = sender
        self.codec = codec
        self.message_processor = message_processor
        self.task_processor = task_processor

    def run(self):
        loop = asyncio.new_event_loop()  # loop = asyncio.get_event_loop()
        loop.run_until_complete(
            send_acl_message(
                self.dest,
                self.illocution,
                self.content,
                self.reply_to,
                self.codec,
                self.message_processor,
                self.task_processor,
            )
        )
        loop.close()


def spawn_send_acl_message(
    dest: AgentCard | str,
    illocution: str,
    content: str,
    sender: str,
    codec: str,
    message_processor=default_message_handler,
    task_processor=default_task_handler,
) -> None:
    t = SendMessageThread(
        dest, illocution, content, sender, codec, message_processor, task_processor
    )
    t.start()
    print("(Send Thread launched.)")



from dataclasses import dataclass
from enum import StrEnum


class Illocution(StrEnum):
    """Illocutions that are used in messages."""

    TELL = "tell"
    ACHIEVE = "achieve"
    ASK = "ask"


@dataclass
class ACLMessage:
    illocution: Illocution
    content: str
    sender: str
    codec: str
    task_id: str | None = None



from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.types import AgentCard

import a2a_acl.content_codecs.common as common
from a2a_acl.a2a_acl_protocol.message_codec import bdi_of_a2a
from a2a_acl.a2a_acl_protocol.acl_message import Illocution, ACLMessage
from a2a_acl.a2a_acl_protocol.send_acl_message import spawn_send_acl_message
from a2a_acl.content_codecs.common import Codec
from a2a_acl.acl_interface.interface import ACLAgentCard

import a2a_acl.content_codecs.python_codec as python_codec
import a2a_acl.content_codecs.atom_codec as atom_codec



class ACLAgentExecutor(AgentExecutor):

    def __init__(self, agentcard: ACLAgentCard, my_url: str):
        self.public_entry_points = [s.functor for s in agentcard.skills]
        self.codecs = agentcard.supported_codecs
        self.print_entry_points()
        self.print_codecs()
        self.card = agentcard
        self.my_url = my_url
        self.codec_objects  = { common.python_codec_id : python_codec.codec_object , common.atom_codec_id : atom_codec.codec_object }

    def print_entry_points(self):
        print("Public entry points:")
        for entry_point in self.public_entry_points:
            print(" " + str(entry_point))

    def print_codecs(self):
        print("Public codecs:")
        for codec in self.codecs:
            print(" " + str(codec))

    def get_codec_object(self, codec_name: str) -> Codec:
        if codec_name in self.codec_objects :
            return self.codec_objects[codec_name]
        else:
            print("Codec " + codec_name + " not supported.")
            raise NotImplementedError

    def firewall_accept(self, m: ACLMessage) -> bool:
        c = self.get_codec_object(m.codec)
        ep = c.extract_entry_point(m.content)
        if ep not in self.public_entry_points:
            print("Received a message on an entry point which is not public: " + ep)
            print("Message not handled.")
            return False
        else:
            return True

    def codec_accept(self, m: ACLMessage) -> bool:
        if m.codec in self.codecs:
            return True
        else:
            print("Codec not supported : " + m.codec)
            print("Message not handled.")
            return False

    async def execute(self, context: RequestContext, event_queue: EventQueue) -> None:
        m: ACLMessage = bdi_of_a2a(context)
        if self.codec_accept(m) and self.firewall_accept(m):
            match m.illocution:
                case Illocution.TELL:
                    await self.execute_tell(m, event_queue)
                case Illocution.ACHIEVE:
                    await self.execute_achieve(m, event_queue)
                case Illocution.ASK:
                    await self.execute_ask(m, event_queue)

    # Override-me
    async def execute_tell(self, m: ACLMessage, event_queue: EventQueue) -> None:
        pass

    # Override-me
    async def execute_achieve(self, m: ACLMessage, event_queue: EventQueue) -> None:
        pass

    # Override-me
    async def execute_ask(self, m: ACLMessage, event_queue: EventQueue) -> None:
        pass

    def check_public_entry_point(self, m: ACLMessage, codec: Codec):
        entry_point = codec.extract_entry_point(m.content)
        return entry_point in self.public_entry_points

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        raise Exception("Cancel not supported.")

    def send_message(
        self, dest: AgentCard | str, illocution: str, content: str, codec: str
    ):
        spawn_send_acl_message(dest, illocution, content, self.my_url, codec)



from uuid import uuid4

import httpx
from a2a.client import A2AClient, A2AClientTimeoutError
from a2a.server.events import EventQueue
from a2a.types import (
    SendMessageRequest,
    MessageSendConfiguration,
    PushNotificationConfig,
    MessageSendParams,
    AgentCard,
)
from a2a.utils import new_agent_text_message

from a2a_acl.a2a_acl_protocol.message_codec import extract_text


async def send_a2a_request(
    client: A2AClient, request: SendMessageRequest
) -> str | None:
    try:
        response = await client.send_message(request)
        t = extract_text(response)
        print("(Synchronous reply received: " + str(t) + ")")
        return t
    except A2AClientTimeoutError:
        print("Warning: No acknowledgement received before timeout.")
        return None


async def send_a2a_message(dest: AgentCard | str, request: SendMessageRequest):
    async with httpx.AsyncClient(timeout=httpx.Timeout(timeout=30)) as httpx_client:
        if isinstance(dest, AgentCard):
            client = A2AClient(httpx_client=httpx_client, agent_card=dest)
        else:
            client = A2AClient(httpx_client=httpx_client, url=dest)

        res = await send_a2a_request(client, request)
        return res


async def sync_reply(output_event_queue: EventQueue, r: str):
    await output_event_queue.enqueue_event(new_agent_text_message(r))


def build_basic_text_request(content: str, reply_to_url: str) -> SendMessageRequest:
    c = MessageSendConfiguration(
        push_notification_config=PushNotificationConfig(url=reply_to_url)
    )
    m = {
        "message": {
            "role": "user",
            "parts": [
                {
                    "kind": "text",
                    "text": content,
                }
            ],
            "messageId": uuid4().hex,
        },
        "configuration": c,
    }
    params = MessageSendParams(**m)
    return SendMessageRequest(id=str(uuid4()), params=params)



from a2a.types import AgentExtension

PROTOCOL_URI = "https://gitlab.eclipse.org/eclipse-research-labs/mosaico-project/a2a-acl/-/blob/main/a2a_acl_protocol/MOSAICO_A2A_ACL_PROTOCOL"
EXTENSION = AgentExtension(
    uri=PROTOCOL_URI, description="MOSAICO A2A ACL", required=True
)



from a2a.types import AgentSkill

from a2a_acl.a2a_acl_protocol.acl_message import Illocution
from a2a_acl.acl_interface.interface import (
    DeclarativeForce,
    SkillDeclaration,
)


def illoc_of_decl(f: DeclarativeForce) -> Illocution:
    match f:
        case DeclarativeForce.BELIEF:
            return Illocution.ASK
        case DeclarativeForce.INPUT:
            return Illocution.TELL
        case DeclarativeForce.ACTION:
            return Illocution.ACHIEVE
        case _:
            raise Exception


def decl_of_illoc(f: Illocution) -> DeclarativeForce:
    match f:
        case Illocution.ASK:
            return DeclarativeForce.BELIEF
        case Illocution.TELL:
            return DeclarativeForce.INPUT
        case Illocution.ACHIEVE:
            return DeclarativeForce.ACTION
        case _:
            raise Exception


def generate_param_example(i: int) -> str:
    return "(" + ("0," * i).removesuffix(",") + ")"


def pretty_print_example(s: SkillDeclaration):
    a = s.arity
    ex_param = "" if a == 0 else generate_param_example(a)
    ex = s.functor + ex_param
    return "(" + str(illoc_of_decl(s.declaration_kind)) + "," + ex + ")"


def parse_example(s: str) -> tuple[str, str]:
    """
    Example of input: '(achieve,share_secret(0))'
    Input encoded by pretty_print_example above.
    """
    s2 = s.removeprefix("(").removesuffix(")")
    (a, b) = s2.split(",")
    f = b.split("(")[0]
    return a, f


def encode_arity(a: int) -> str:
    return "(needs " + str(a) + " parameter" + ("s" if a > 1 else "") + ")"


def decode_arity(s: str) -> int:
    """
    Example of input: 'send a secret (needs 1 parameter)'
    Inputs are encoded by encode_arity above.
    Return 1 in this example.
    """
    l = s.split(" ")
    s1 = l[-2]
    return int(s1)


def a2a_skill_of_bdi_skill(s: SkillDeclaration) -> AgentSkill:
    return AgentSkill(
        id=s.functor,
        name=s.functor + " (from ACL agent)",
        description=s.doc + " " + encode_arity(s.arity),
        tags=[s.functor],
        examples=[pretty_print_example(s)],
    )


def bdi_skill_of_a2a_skill(s: AgentSkill) -> SkillDeclaration:
    """
    Exemple of skill for a ACL agent :
    AgentSkill(
        description='send a secret (needs 1 parameter)',
        examples=['(achieve,share_secret(0))'],
        id='share_secret',
        input_modes=None,
        name='share_secret (from ACL agent)',
        output_modes=None,
        security=None,
        tags=['share_secret']
        )
    """
    (a, f) = parse_example(s.examples[0])
    return SkillDeclaration(
        declaration_kind=decl_of_illoc(Illocution(a)),
        functor=f,
        arity=decode_arity(s.description),
        doc=s.description,
    )
